@@ -10,6 +10,7 @@ import dill
 import grpc
 import isolate
 import yaml
+from isolate.backends.common import active_python
 from isolate.backends.settings import DEFAULT_SETTINGS
 from isolate.connections import PythonIPC
 from koldstart.sdk import (
@@ -23,6 +24,8 @@ from koldstart.sdk import (
     get_default_credentials,
 )
 
+dill.settings["recurse"] = True
+
 ReturnT = TypeVar("ReturnT")
 BasicConfig = Dict[str, Any]
 _UNSET = object()
@@ -80,6 +83,20 @@ class Host:
         raise NotImplementedError
 
 
+def _execution_controller(
+    func: Callable[..., ReturnT],
+    args: tuple[Any, ...],
+    kwargs: dict[str, Any],
+) -> Callable[..., ReturnT]:
+    """Handle the execution of the given user function."""
+
+    @wraps(func)
+    def wrapper(*remote_args: Any, **remote_kwargs: Any) -> ReturnT:
+        return func(*remote_args, *args, **remote_kwargs, **kwargs)
+
+    return wrapper
+
+
 @dataclass
 class LocalHost(Host):
     # The environment which provides the default set of
@@ -143,20 +160,27 @@ class KoldstartHost(Host):
         args: tuple[Any, ...],
         kwargs: dict[str, Any],
     ) -> ReturnT:
-        environments = [self._connection.define_environment(**options.environment)]
+        environment_options = options.environment.copy()
+        environment_options.setdefault("python_version", active_python())
+        environments = [self._connection.define_environment(**environment_options)]
 
-        machine_type = options.host.pop("machine_type", KOLDSTART_DEFAULT_MACHINE_TYPE)
-        keep_alive = options.host.pop("keep_alive", KOLDSTART_DEFAULT_KEEP_ALIVE)
+        machine_type = options.host.get("machine_type", KOLDSTART_DEFAULT_MACHINE_TYPE)
+        keep_alive = options.host.get("keep_alive", KOLDSTART_DEFAULT_KEEP_ALIVE)
+        setup_function = options.host.get("setup_function", None)
 
         machine_requirements = MachineRequirements(
             machine_type=machine_type, keep_alive=keep_alive
         )
 
         return_value = _UNSET
+        # Allow isolate provided arguments (such as setup function) to take
+        # precedence over the ones provided by the user.
+        partial_func = _execution_controller(func, args, kwargs)
         for partial_result in self._connection.run(
-            partial(func, *args, **kwargs),
+            partial_func,
             environments,
             machine_requirements=machine_requirements,
+            setup_function=setup_function,
         ):
             for log in partial_result.logs:
                 print(str(log))
